/* 
 * Copyright (C) Shivaram Upadhyayula <shivaram.u@quadstor.com>
 */

#ifndef QS_COMMONDEFS_H_
#define QS_COMMONDEFS_H_ 1

#define QUADSTOR_VIRT_VERSION	"3.2.30"

#define TL_MAX_DISKS	512
#define TL_MAX_POOLS	512	
#define TL_MAX_DEVICES		4096
#define TL_MAX_TDISKS		TL_MAX_DEVICES
#define MAX_DINFO_DEVICES	TL_MAX_DEVICES
#define TL_DEV_NAME	"iodev"
#ifdef ENABLE_FCINT
#define FC_DEV_NAME	"qlascdev"
#define FC_DEV		"/dev/qlascdev"
#else
#define FC_DEV_NAME	"ql2xapidev"
#define FC_DEV		"/dev/ql2xapidev"
#endif
#define TL_DEV		"/dev/iodev"
#define MIN_PHYSDISK_SIZE	(1ULL << 32)
#define MDAEMON_USR_PATH	"/var/run/vtmdaemonusr.sock"
#define CDAEMON_RECV_PATH	"/var/run/vtcdaemonrcv.sock"


/* Limits */

#define DELETE_MAGIC			0x5a4A3a2A
#define TL_RID_MAX	40
struct mdaemon_info {
	int cmd_debug;
	int access_debug;
	int cluster_debug;
	int load_debug;
	int debug_unmap;
	int extended_logging;
	int reservation_debug;
	int dedupe_threshold;
	int rcache_threshold;
	int amap_threshold;
	int index_threshold;
	int ext_dedupe_mem;
	int flush_dedupe_mem;
	int amap_mirror_threads; 
	int node_sock_timeout;
	int ignore_src_errors;
	int ignore_unaligned_uncomp_errors;
	int sequential_tables;
	int debug_meta;
	int sock_read_timeout;
	int meta_load_timeout;
	int meta_load_retries;
	int rep_sync_sock;
	uint32_t ddtable_sync_low;
	uint32_t ddtable_sync_high;
	uint32_t ddtable_sync_pause;
	uint32_t ddtable_sync_debug;
};

#define UMA_INFO_BUFFER_SIZE	65536
struct uma_info {
	char name[32];
	uint64_t objs;
	uint64_t alloced;
	uint32_t umabufs;
	uint32_t free_umabufs;
	uint32_t size;
	uint32_t objs_per_umabuf;
	uint32_t umabuf_size;
};

struct uma_spec {
	uint8_t *buffer;
};

#define TDISK_NAME_LEN		128
#define TDISK_MAX_NAME_LEN	136
#define TDISK_MIN_NAME_LEN	40
#define TL_NAME_LEN		36
#define TL_MAX_NAME_LEN		40
#define GROUP_NAME_LEN		32
#define GROUP_MAX_NAME_LEN	40
#define TL_MAX_ISCSI_NAME_LEN	256
#define TL_ISCSI_NAME_LEN	(TL_MAX_ISCSI_NAME_LEN  - 1)

struct iscsichap {
	int dir;
	uint32_t target_id;
	char user[TL_MAX_ISCSI_NAME_LEN];
	char passwd[TL_MAX_ISCSI_NAME_LEN];
	char iqn[TL_MAX_ISCSI_NAME_LEN];
	TAILQ_ENTRY(iscsichap) c_entry;
};
TAILQ_HEAD(chaplist, iscsichap);


struct iscsiconf {
	uint32_t target_id;
	char iqn[256];
	struct chaplist chaplist;
};

struct tdisk_stats {
	uint64_t write_size;
	uint64_t read_size;
	uint64_t unaligned_size;
	uint64_t blocks_deduped;
	uint64_t inline_deduped;
	uint64_t post_deduped;
	uint64_t zero_blocks;
	uint64_t unmap_blocks;
	uint64_t uncompressed_size;
	uint64_t compressed_size;
	uint64_t compression_hits;
	uint64_t compression_misses;
	uint64_t verify_hits;
	uint64_t verify_misses;
	uint64_t verify_errors;
	uint64_t inline_waits;
	uint64_t cw_hits;
	uint64_t cw_misses;
	uint64_t xcopy_write;
	uint64_t xcopy_read;
	uint64_t wsame_blocks;
	uint64_t write_cmds;
	uint64_t read_cmds;
	uint64_t unmap_cmds;
	uint64_t wsame_cmds;
	uint64_t xcopy_cmds;
	uint64_t populate_token_cmds;
	uint64_t write_using_token_cmds;
	uint64_t populate_token_size;
	uint64_t write_using_token_size;
	uint64_t write_ticks;
	uint64_t read_ticks;
	uint64_t unmap_ticks;
	uint64_t wsame_ticks;
	uint64_t xcopy_read_ticks;
	uint64_t xcopy_write_ticks;
	uint64_t populate_token_ticks;
	uint64_t write_using_token_ticks;
	uint64_t sync_cache_cmds;
	uint64_t merged_write_size;
	uint64_t merged_write_cmds;
	uint64_t pad[43];
} __attribute__ ((__packed__));

struct tdrive_stats {
	uint64_t write_cmds;
	uint64_t read_cmds;
	uint64_t read_cmd_ticks;
	uint64_t read_kernel_ticks;
	uint64_t read_usr_ticks;
	uint64_t read_interface_ticks;
	uint64_t read_dxfer_ticks;
	uint64_t write_cmd_ticks;
	uint64_t write_kernel_ticks;
	uint64_t write_usr_ticks;
	uint64_t write_interface_ticks;
	uint64_t write_allocate_buffers_ticks;
	uint64_t write_dxfer_ticks;
	uint64_t write_bytes_processed;
	uint64_t read_bytes_processed;
	uint64_t ctio_mmap_alloc;
	uint64_t ctio_mmap_alloc_cmds;
	uint64_t ctio_nonmmap_alloc;
	uint64_t ctio_nonmmap_alloc_cmds;
	uint64_t ctio_saved_alloc;
	uint64_t ctio_saved_alloc_cmds;
};

struct mirror_state {
	uint8_t mirror_type;
	uint8_t mirror_role;
	uint8_t next_role;
	uint8_t prev_role;
	uint16_t mirror_target_id;
	uint8_t percent;
	uint8_t pad2;
	int mirror_flags;
	uint32_t pad3;
	uint32_t mirror_src_ipaddr;
	uint32_t mirror_ipaddr;
	char mirror_vdisk[40];
	char mirror_group[GROUP_MAX_NAME_LEN];
	uint64_t pad4;
	uint64_t pad5;
	char sys_rid[40];
	char mirror_vdisk_ext[96];
};

enum {
	VDISK_DELETED = 0x1,
	VDISK_DELETING = 0x2,
};

struct tdisk_info {
	uint64_t size;
	char name[TDISK_MAX_NAME_LEN];
	char group_name[GROUP_MAX_NAME_LEN];
	char serialnumber[36];
	struct mirror_state mirror_state;
	uint32_t target_id;
	int free_alloc;
	uint32_t tl_id;
	int iscsi_tid;
	int vhba_id;
	uint8_t online;
	uint8_t disabled;
	uint8_t enable_deduplication;
	uint8_t enable_compression;
	uint8_t enable_verify;
	uint8_t force_inline;
	uint8_t lba_shift;
	uint8_t clone_error;
	uint8_t mirror_error;
	int8_t delete_error;
	uint8_t attach;
	uint8_t v2_format;
	uint8_t threshold;
	uint32_t group_id;
	struct iscsiconf iscsiconf;
	struct tdisk_stats stats;
	struct tdrive_stats tstats;
	TAILQ_ENTRY(tdisk_info) g_entry;
	TAILQ_ENTRY(tdisk_info) q_entry;
	struct group_info *group;
};

enum {
	TDISK_ENABLE_DEDUPLICATION = 1,
	TDISK_DISABLE_DEDUPLICATION,
	TDISK_ENABLE_COMPRESSION,
	TDISK_DISABLE_COMPRESION,
};

#define MAX_TARGET_SIZE		(1ULL << 46) /* 64 TB */

struct bint_stats {
	uint64_t dedupe_blocks;
	uint64_t compressed_size;
	uint64_t uncompressed_size;
	uint64_t total_blocks;
	uint64_t compression_hits;
	uint64_t pad2;
	uint64_t pad3;
	uint64_t pad4;
	uint64_t pad5;
	uint64_t pad6;
	uint64_t pad7;
	uint64_t pad8;
	uint64_t pad9;
	uint64_t pad10;
};

#define V2_DISK		0x1
#define V2_LOG_FORMAT	0x2
#define RID_SET		0x4
#define V3_LOG_FORMAT	0x8

struct raw_bdevint {
	uint8_t ddmaster;
	uint8_t ddbits;
	uint8_t log_disk;
	uint8_t log_disks;
	int8_t initialized;
	uint8_t enable_comp;
	uint8_t log_write;
	uint8_t flags;
	uint32_t bid;
	uint32_t sector_shift;
	uint64_t usize;
	uint64_t free;
	uint64_t b_start;
	uint64_t b_end;
	uint8_t magic[8];
	uint8_t vendor[8];
	uint8_t product[16];
	uint8_t serialnumber[32];
	uint32_t group_id;
	int32_t group_flags;
	uint8_t write_cache;
	uint8_t quad_prod[4];
	uint8_t dismagic[3];
	uint64_t pad3;
	uint64_t pad4;
	uint64_t pad5;
	uint64_t pad6;
	uint64_t pad7;
	uint64_t pad8;
	uint64_t pad9;
	uint64_t pad10;
	struct bint_stats stats;
	char mrid[TL_RID_MAX];
	char group_name[GROUP_MAX_NAME_LEN];
	uint8_t ext_serialnumber[224];
} __attribute__ ((__packed__));

static inline int
raw_bint_serial_match(struct raw_bdevint *raw_bint, char *serialnumber, int serial_len)
{
	int serial_max, min_len;

	min_len = serial_max = sizeof(raw_bint->serialnumber);
	if (min_len > serial_len)
		min_len = serial_len;

	if (memcmp(raw_bint->serialnumber, serialnumber, min_len))
		return 0;

	if (serial_len > serial_max) {
		if (memcmp(raw_bint->ext_serialnumber, serialnumber + serial_max, serial_len - serial_max))
			return 0;
	}
	return 1;
}

struct group_conf {
	char name[GROUP_MAX_NAME_LEN];
	uint32_t group_id;
	int dedupe;
	int dedupemeta;
	int dedupelite;
};

#define DEFAULT_GROUP_NAME	"Default"

struct bdev_info {
	uint32_t bid;
	char devpath[256];
	uint64_t size;
	uint64_t usize;
	uint64_t free;
	uint64_t reserved;
	uint8_t  isnew;
	uint8_t  reopen;
	uint8_t  ddmaster;
	uint8_t  ismaster;
	uint8_t  log_disk;
	uint8_t  ha_disk;
	uint8_t  unmap;
	uint8_t  write_cache;
	uint8_t load_post;
	uint8_t first_load;
	int8_t initialized;
	uint8_t enable_comp;
	uint8_t  vendor[8];
	uint8_t  product[16];
	uint8_t  serialnumber[256];
	uint32_t free_alloc;
	int serial_len;
	int use_diskid;
	uint32_t max_index_groups;
	uint32_t group_id;
	struct bint_stats stats;
	char rid[TL_RID_MAX];
	char errmsg[256];
};

#define DISK_INDEX_MAGIC	"QSDINDEX"
#define INDEX_TABLE_PAD_MAX	74

struct raw_index_data {
	uint8_t magic[8];
	uint8_t name[40];
	uint8_t serialnumber[32];
	uint64_t size;
	uint16_t target_id;
	uint16_t lba_shift;
	uint32_t flags;
	uint32_t group_id;
	uint8_t  threshold;
	uint8_t pad3[3];
	uint64_t max_size;
	uint64_t pad5;
	uint64_t pad6;
	uint64_t pad7;
	uint64_t pad8;
	uint64_t pad9;
	uint64_t pad10;
	uint64_t table_pad[INDEX_TABLE_PAD_MAX]; 
	struct tdisk_stats stats;
	struct mirror_state mirror_state;
	uint8_t ext_name[96];
};

struct node_config {
	int node_type;
	char node_host[16];
	char controller_host[16];
	char recv_host[16];
	char ha_host[16];
	char ha_bind_host[16];
	uint32_t nodes[64];
	uint32_t node_ipaddr;
	uint32_t controller_ipaddr;
	uint32_t recv_ipaddr;
	uint32_t ha_ipaddr;
	uint32_t ha_bind_ipaddr;
	int sync_status;
	int node_flags;
	int node_role;
	int recv_flags;
	int fence_enabled;
	uint16_t mirror_recv_timeout;
	uint16_t mirror_send_timeout;
	uint16_t mirror_sync_timeout;
	uint16_t mirror_sync_recv_timeout;
	uint16_t mirror_sync_send_timeout;
	uint16_t client_send_timeout;
	uint16_t controller_recv_timeout;
	uint16_t node_sync_timeout;
	uint16_t ha_check_timeout;
	uint16_t ha_ping_timeout;
	uint32_t client_connect_timeout;
	uint32_t mirror_connect_timeout;
};

enum {
	OP_CLONE,
	OP_MIRROR,
};

enum {
	CLONE_STATUS_SUCCESSFUL = 0x1,
	CLONE_STATUS_ERROR,
	CLONE_STATUS_INPROGRESS,
};

enum {
	MIRROR_STATUS_SUCCESSFUL = CLONE_STATUS_SUCCESSFUL,
	MIRROR_STATUS_ERROR = CLONE_STATUS_ERROR,
	MIRROR_STATUS_INPROGRESS = CLONE_STATUS_INPROGRESS,
};

struct job_stats {
	uint64_t elapsed_msecs;
	uint64_t read_msecs;
	uint64_t write_msecs;
	uint64_t hash_compute_msecs;
	uint64_t hash_lookup_msecs;
	uint32_t dest_ipaddr;
	uint32_t src_ipaddr;
	uint64_t mapped_blocks;
	uint64_t deduped_blocks;
	uint64_t refed_blocks;
	uint64_t bytes_read;
	uint64_t blocks_read;
	uint64_t blocks_written;
	uint64_t bytes_written;
};

struct clone_config {
	uint32_t src_target_id;
	uint32_t dest_target_id;
	uint32_t dest_ipaddr;
	uint32_t src_ipaddr;
	uint64_t job_id;
	uint8_t  op;
	uint8_t  status;
	uint8_t  progress;
	uint8_t  attach;
	uint8_t  mirror_type;
	uint8_t  mirror_role;
	char mirror_vdisk[TDISK_MAX_NAME_LEN];
	char mirror_group[GROUP_MAX_NAME_LEN];
	char sys_rid[40];
	char errmsg[256];
	struct job_stats stats;
};

enum {
	GROUP_FLAGS_MASTER,
	GROUP_FLAGS_HA_DISK,
	GROUP_FLAGS_DEDUPEMETA,
	GROUP_FLAGS_LOGDATA,
	GROUP_FLAGS_UNMAP_ENABLED,
	GROUP_FLAGS_UNMAP,
	GROUP_FLAGS_TAIL_META,
	GROUP_FLAGS_SKIP_ALLOC_LIST,
	GROUP_FLAGS_STATS_BLOCK,
	GROUP_FLAGS_VARDEDUPE,
	GROUP_FLAGS_DEDUPE,
	GROUP_FLAGS_DEDUPE_DISABLED,
	GROUP_FLAGS_VARDEDUPE3,
	GROUP_FLAGS_DEDUPE_LITE,
};

enum {
	MIRROR_TYPE_NONE,
	MIRROR_TYPE_ACTIVE,
	MIRROR_TYPE_PASSIVE,
};

enum {
	MIRROR_ROLE_UNKNOWN,
	MIRROR_ROLE_MASTER,
	MIRROR_ROLE_PEER,
};

static inline char *
mirror_role_str(int role)
{
	switch (role) {
	case MIRROR_ROLE_MASTER:
		return "Master";
	case MIRROR_ROLE_PEER:
		return "Slave";
	default:
		return "Unknown";
	}
}

struct usr_msg {
	int msg_id;
	int msg_rsp;
	int iscsi_tid;
	int vhba_id;
	uint32_t target_id;
	uint32_t mirror_ipaddr;
	uint32_t pad;
};

struct usr_job {
	struct usr_msg msg;
	uint64_t job_id;
	struct job_stats stats;
};

struct usr_notify {
	struct usr_msg msg;
	int notify_type;
	int pad;
	char notify_msg[96];
};

enum {
	USR_NOTIFY_INFO,
	USR_NOTIFY_WARN,
	USR_NOTIFY_ERR,
	USR_NOTIFY_VDISK_THRESHOLD,
	USR_NOTIFY_BINT_CREATE_DONE,
};

enum {
	USR_RSP_OK,
	USR_RSP_ERR,
	USR_RSP_MIRROR_ACTIVE,
	USR_RSP_FENCE_FAILED,
	USR_RSP_FENCE_SUCCESSFUL,
	USR_RSP_BID_INVALID,
	USR_RSP_FENCE_MANUAL,
};

enum {
	USR_MSG_MIRROR_CHECK,
	USR_MSG_ATTACH_INTERFACE,
	USR_MSG_JOB_COMPLETED,
	USR_MSG_VDISK_DELETED,
	USR_MSG_FENCE_NODE,
	USR_MSG_BID_VALID,
	USR_MSG_NOTIFY,
};

struct fc_rule_config {
	uint64_t wwpn[2];
	uint64_t target_wwpn;
	uint32_t target_id;
	uint64_t old_wwpn;
	int rule;
};

enum {
	FC_RULE_ALLOW,
	FC_RULE_DISALLOW,
};

#define RCACHE_THRESHOLD_MIN			5
#define RCACHE_THRESHOLD_MAX			35
#define AMAP_THRESHOLD_MIN			2
#define AMAP_THRESHOLD_DEFAULT			6
#define AMAP_THRESHOLD_EXT			5
#define AMAP_THRESHOLD_MAX			12
#define INDEX_THRESHOLD_MIN			4
#define INDEX_THRESHOLD_DEFAULT			8
#define INDEX_THRESHOLD_EXT			6
#define INDEX_THRESHOLD_MAX			15
#define DEDUPE_THRESHOLD_MIN			56
#define DEDUPE_THRESHOLD_MAX			75

#define MIRROR_RECV_TIMEOUT_MAX			300	
#define MIRROR_CONNECT_TIMEOUT_MAX		120
#define MIRROR_SEND_TIMEOUT_MAX			300	
#define MIRROR_SYNC_TIMEOUT_MAX			300	
#define MIRROR_SYNC_RECV_TIMEOUT_MAX		300	
#define MIRROR_SYNC_SEND_TIMEOUT_MAX		300	
#define CLIENT_SEND_TIMEOUT_MAX			300	
#define CLIENT_CONNECT_TIMEOUT_MAX		300	
#define CONTROLLER_RECV_TIMEOUT_MAX		90	
#define CONTROLLER_CONNECT_TIMEOUT_MAX		70
#define NODE_SYNC_TIMEOUT_MAX			90
#define AMAP_MIRROR_THREADS_MAX			16
#define HA_CHECK_TIMEOUT_MAX			60
#define HA_PING_TIMEOUT_MAX			60
#define NODE_SOCK_TIMEOUT_MAX			120

#define MIRROR_RECV_TIMEOUT_MIN			20
#define MIRROR_CONNECT_TIMEOUT_MIN		5
#define MIRROR_SEND_TIMEOUT_MIN			20
#define MIRROR_SYNC_TIMEOUT_MIN			15	
#define MIRROR_SYNC_RECV_TIMEOUT_MIN		15
#define MIRROR_SYNC_SEND_TIMEOUT_MIN		15
#define CLIENT_SEND_TIMEOUT_MIN			15
#define CLIENT_CONNECT_TIMEOUT_MIN		5
#define CONTROLLER_RECV_TIMEOUT_MIN		15
#define CONTROLLER_CONNECT_TIMEOUT_MIN		5
#define NODE_SYNC_TIMEOUT_MIN			20
#define AMAP_MIRROR_THREADS_MIN			1
#define HA_CHECK_TIMEOUT_MIN			5
#define HA_PING_TIMEOUT_MIN			2
#define NODE_SOCK_TIMEOUT_MIN			10

enum {
	WRITE_CACHE_DEFAULT,
	WRITE_CACHE_FLUSH,
	WRITE_CACHE_FUA,
	WRITE_CACHE_BBU,
};

enum {
	MIRROR_FLAGS_CONFIGURED,
	MIRROR_FLAGS_DISABLED,
	MIRROR_FLAGS_NEED_RESYNC,
	MIRROR_FLAGS_PEER_LOAD_DONE,
	MIRROR_FLAGS_STATE_INVALID,
	MIRROR_FLAGS_IN_RESYNC,
	MIRROR_FLAGS_BLOCK,
	MIRROR_FLAGS_WAIT_FOR_MASTER,
	MIRROR_FLAGS_WAIT_FOR_PEER,
	MIRROR_FLAGS_WRITE_BITMAP_VALID,
	MIRROR_FLAGS_IN_SHUTDOWN,
};

#define TDISK_UNMAP_LBA_COUNT				512 /* 2 MB */
#define TDISK_UNMAP_LBA_COUNT_LEGACY			4096 /* 2 MB */
#define FC_MAX_VISIBLE_LUN				255

struct hba_info {
	int host;
	int enable;
	int debug;
	uint64_t wwpn;
};

#define FC_INITIATOR_MODE		0x01
#define FC_TARGET_MODE			0x02
#define FC_DUAL_MODE			(FC_INITIATOR_MODE | FC_TARGET_MODE)
static inline char *
get_hba_mode_str(int mode)
{
	switch (mode) {
	case FC_INITIATOR_MODE:
		return "initiator";
	case FC_TARGET_MODE:
		return "target";
	case FC_DUAL_MODE:
		return "initiator|target";
	default:
		return "unknown";
	}
}

struct ldev_info {
	uint32_t tid;
	uint32_t lun_id;
	uint32_t iscsi_tid;
	uint32_t vhba_id;
	uint32_t lba_shift;
};

enum {
	NODE_STATUS_OK			= SCSI_STATUS_OK,
	NODE_STATUS_SCSI_SENSE		= SCSI_STATUS_CHECK_COND,
	NODE_STATUS_BUSY		= SCSI_STATUS_BUSY,
	NODE_STATUS_RESERV_CONFLICT	= SCSI_STATUS_RESERV_CONFLICT,

	/* Custom codes */
	NODE_STATUS_TARGET_NOT_FOUND	= 0xE0,
	NODE_STATUS_INVALID_MSG		= 0xE1,
	NODE_STATUS_UNREGISTERED_NODE	= 0xE2,
	NODE_STATUS_MEM_ALLOC_FAILURE	= 0xE3,
	NODE_STATUS_ERROR		= 0xE4,
	NODE_STATUS_AMAP_NOT_FOUND	= 0xE5,
	NODE_STATUS_AMAP_NEEDS_SYNC	= 0xE6,
	NODE_STATUS_INITIALIZING	= 0xE7,
	NODE_STATUS_NEED_RESYNC		= 0xE8,
	NODE_STATUS_MIRROR_CONFIGURED	= 0xE9,
	NODE_STATUS_INVALID_MIRROR_CONFIGURATION	= 0xEA,
	NODE_STATUS_SKIP_LOCAL_WRITE	= 0xEB,
	NODE_STATUS_IS_MASTER		= 0xEC,
	NODE_STATUS_DO_LOCAL_READ	= 0xED,
	NODE_STATUS_OK_V4		= 0xEF,
};

struct core_node_msg {
	uint32_t csum;
	uint32_t dxfer_len;
	uint8_t  msg_cmd;
	uint8_t  msg_status;
	uint16_t pad;
	uint32_t pad1;
	uint64_t msg_id;
	uint8_t  data[0];
};

static inline void
calc_sector_bits(uint32_t sector_size, uint32_t *sector_shift)
{
	uint32_t shift;

	shift = 0;
	while ((sector_size >>= 1))
	{
		shift++;
	}
	*sector_shift = shift;
	return;
}

struct page_spec {
	unsigned long addr;
};

struct pgdata_spec {
	unsigned long addr;
	uint16_t pg_len;
	uint16_t pg_offset;
	uint32_t pad;
};

#define IOCTL_PGADDR_MAX	256
struct bio_spec {
	unsigned long kfd;
	uint8_t rw;
	uint16_t page_count;
	uint32_t length;
	uint64_t b_start;
	unsigned long bio_addr;
	struct pgdata_spec *pglist;
};

struct comp_spec {
	uint16_t cpage_count;
	uint16_t upage_count;
	uint16_t compress;
	uint16_t pad;
	uint32_t comp_size;
	uint32_t uncomp_size;
	struct page_spec *cpages;
};

struct file_spec {
	unsigned long kfd;
	unsigned long length;
	int error;
	int isnew;
	char name[4096];
};

struct bdev_sync {
	unsigned long kfd;
};

struct tdev_spec {
	unsigned long kfd;
	int error;
	uint32_t host;
	uint32_t channel;
	uint32_t target;
	uint64_t lun;
	char name[1024];
};

struct bdev_spec {
	unsigned long kfd;
	int error;
	int pad;
	uint32_t sector_shift;
	uint32_t sector_size;
	uint64_t size;
	char name[1024];
};

struct ctio_mmap_setup {
	unsigned long addr;
	uint32_t lun_id;
	uint32_t pad;
};

struct bio_mmap_setup {
	unsigned long addr;
};

struct ctio_wakeup {
	uint32_t lun_id;
	uint32_t pad;
};

static inline uint32_t
csum32(uint8_t *buf, int len)
{
        int i;
        uint32_t csum = 0;

        for (i = 0; (i + 4) <= len ; i+=4)
        {
                uint32_t val = *((uint32_t *)(buf+i));
                csum += val;
        }
	while (i < len)
		csum += buf[i++];
        return ~(csum);
}

enum {
	COMP_ALG_LZF	= 0x00,
	COMP_ALG_LZ4	= 0x01,
};

#define COMP_ALG_SHIFT	14
#define COMP_ALG_MASK	((1 << COMP_ALG_SHIFT) - 1)
#define SET_COMP_SIZE(ptr, csize, alg)	(*((uint16_t *)(ptr)) = (csize | (alg << COMP_ALG_SHIFT)))

#ifndef LBA_SHIFT
#define LBA_SHIFT		12
#define LBA_SIZE		(1U << LBA_SHIFT)
#define LBA_MASK		(LBA_SIZE - 1)
#define BDEV_META_RESERVED	262144
#define BDEV_META_OFFSET	BDEV_META_RESERVED
#define BDEV_STATS_OFFSET	(BDEV_META_OFFSET - 32768)
#endif

#ifndef WRITE_6
#define	WRITE_6			0x0A
#endif
#ifndef WRITE_10
#define	WRITE_10		0x2A
#endif
#ifndef WRITE_12
#define	WRITE_12		0xAA
#endif
#ifndef WRITE_16
#define	WRITE_16		0x8A
#endif
#ifndef READ_16
#define	READ_16			0x88
#endif
#ifndef UNMAP
#define UNMAP				0x42
#endif
#ifndef WRITE_SAME
#define WRITE_SAME			0x41
#endif
#ifndef WRITE_SAME_16
#define WRITE_SAME_16			0x93
#endif
#ifndef EXTENDED_COPY
#define EXTENDED_COPY			0x83
#endif
#ifndef COMPARE_AND_WRITE
#define COMPARE_AND_WRITE		0x89
#endif
#ifndef MODE_SELECT_6
#define	MODE_SELECT_6		0x15
#endif
#ifndef MODE_SELECT_10
#define	MODE_SELECT_10		0x55
#endif
#ifndef WRITE_ATTRIBUTE
#define WRITE_ATTRIBUTE			0x8D
#endif
#ifndef WRITE_DYNAMIC_RUNTIME_ATTRIBUTE
#define WRITE_DYNAMIC_RUNTIME_ATTRIBUTE		0xD2
#endif
#ifndef SEND_DIAGNOSTIC
#define	SEND_DIAGNOSTIC		0x1D
#endif
#ifndef SECURITY_PROTOCOL_OUT
#define SECURITY_PROTOCOL_OUT		0xB5
#endif
#ifndef SPACE_16
#define SPACE_16			0x91
#endif
#ifndef PERSISTENT_RESERVE_OUT
#define PERSISTENT_RESERVE_OUT		0x5f
#endif
static inline int
is_ctio_write(uint8_t cdb)
{
	switch(cdb) {
	case WRITE_6:
	case WRITE_10:
	case WRITE_12:
	case WRITE_16:
	case WRITE_SAME:
	case WRITE_SAME_16:
	case EXTENDED_COPY:
	case COMPARE_AND_WRITE:
		return 1;
	default:
		return 0;
	}
}

static inline int
is_ctio_read(uint8_t cdb)
{
	switch(cdb) {
	case READ_6:
	case READ_10:
	case READ_12:
	case READ_16:
		return 1;
	default:
		return 0;
	}
}

static inline void
calc_num_blocks(int ctio_dxfer_len, int *ret_pglist_cnt, int *ret_dxfer_len)
{
	int bits = 0, orig_dxfer_len = ctio_dxfer_len;
	int dxfer_len;

	while (ctio_dxfer_len > 1) {
		bits++;
		ctio_dxfer_len >>= 1;
	}

	dxfer_len = (1U << bits);
	if (dxfer_len != orig_dxfer_len)
		dxfer_len <<= 1;

	*ret_pglist_cnt = (dxfer_len >> LBA_SHIFT);
	if (dxfer_len & LBA_MASK)
		*(ret_pglist_cnt) += 1;
	*ret_dxfer_len = dxfer_len;
}

static inline uint64_t
align_size(uint64_t size, int boundary)
{
	uint64_t adjust;

	adjust = (size + (boundary - 1)) & -(boundary);
	return (adjust);
}

#define BIO_PAGE_INCREMENT	128

#endif /* COMMONDEFS_H_ */
