#ifndef QS_COREMSG_H_
#define QS_COREMSG_H_

#define MAP_ALLOCATION_SHIFT	22
#define MAP_ALLOCATION_SIZE	(1U << MAP_ALLOCATION_SHIFT)
#define MIN_ALLOCATION_SIZE	(4096)
#define CTIO_MMAP_SIZE		262144

enum {
	NODE_MSG_CTIO_WRITE_LENGTH,
	NODE_MSG_FC_INITIATOR_CHECK,
	NODE_MSG_FREE_INITIATOR,
	NODE_MSG_DEVICE_RESET,
	NODE_MSG_ABORT_TASK,
	NODE_MSG_ABORT_TASK_SET,
	NODE_MSG_CHECK_CMD,
	NODE_MSG_IOCTL,
	NODE_MSG_CTIO_REQ,
	NODE_MSG_CTIO_WRITE_CMD_ALLOCATE,
	NODE_MSG_CTIO_WRITE_CMD_FREE,
	NODE_MSG_NEW_INITIATOR,
};

struct core_msg_hdr {
	uint8_t  msg_cmd;
	uint8_t  pad[3];
	uint32_t length;
	uint32_t status;
	uint32_t pad2;
};

struct ctio_write_len_req {
	uint64_t lun_id;
	uint8_t cdb[16];
};

struct ctio_write_len_rsp {
	uint32_t block_size;
	uint32_t num_blocks;
	int dxfer_len;
	int pad;
};

struct ctio_free_req {
	uint32_t lun_id;
	uint32_t pad;
	uint64_t tx_id;
};

struct ctio_allocate_req {
	uint8_t cdb[16];
	uint32_t lun_id;
	uint32_t pad;
};

struct ctio_allocate_rsp {
	uint64_t tx_id;
	uint32_t block_size;
	uint32_t num_blocks;
	int dxfer_len;
	int pglist_cnt;
	uint8_t data[0];
};

struct initiator_req {
	uint64_t i_prt[2];
	uint64_t t_prt[2];
	uint64_t lun_id;
	uint32_t init_int;
	uint32_t task_tag;
};

struct new_initiator_req {
	uint64_t i_prt[2];
	uint64_t t_prt[2];
	uint64_t lun_id;
	uint32_t init_int;
	uint32_t r_prt;
	char name[256];
};

struct ioctl_req {
	uint32_t cmd;
	uint32_t pad;
	uint8_t data[0];
};

enum {
	DATA_DIR_NONE,
	DATA_DIR_TO_TAPE,
	DATA_DIR_FROM_TAPE,
};

	
struct ctio_tape_req {
	unsigned long kfd;
	uint8_t cdb[16];
	uint32_t dxfer_len;
	int pglist_cnt;
	int resid;
	int dir;
	uint32_t scsi_status;
	uint32_t sense_len;
	uint8_t sense[64];
	uint8_t *data;
};

struct ctio_req {
	uint8_t cdb[16];
	uint32_t dxfer_len;
	uint32_t task_tag;
	uint32_t init_int;
	uint32_t lun_id;
	uint64_t i_prt[2];
	uint64_t t_prt[2];
	uint32_t r_prt;
	uint32_t pad;
	uint64_t tx_id;
	char init_name[256];
};

struct ctio_rsp {
	uint32_t dxfer_len;
	uint32_t scsi_status;
	uint32_t sense_len;
	uint32_t csum;
	uint32_t pglist_cnt;
	uint32_t ctio_dxfer_len;
	uint64_t tx_id;
	uint8_t data[0];
};

struct ctio_spec {
	int dxfer_len;
	int pglist_cnt;
	uint64_t tx_id;
	uint32_t tid;
	uint32_t pad;
	uint8_t  cdb[16];
	struct pgdata_spec *pglist;
};

struct bio_msg {
	volatile unsigned long bio_addr;
	volatile int error;
	volatile int inuse;
};

struct sock_data_spec {
	int sockfd;
	int page;
	unsigned long addr;
	int offset;
	int len;
	int done;
	int pad;
};

struct sock_local_spec {
	int sockfd;
	int pad;
	char path[512];
};

struct sock_spec {
	int sockfd;
	int acceptfd;
	int linger;
	uint32_t ipaddr;
	uint32_t local_ipaddr;
	uint32_t status;
	uint16_t port;
	uint16_t set;
	uint32_t pad;
};

#define BIO_MSG_SIZE	(512 * 1024)
#define BIO_MAX_MSG	(BIO_MSG_SIZE / sizeof(struct bio_msg))


#endif
