/* 
 * Copyright (C) Shivaram Upadhyayula <shivaram.u@quadstor.com>
 */

#ifndef QS_CLSTDEFS_H_
#define QS_CLSTDEFS_H_

#include "itfdefs.h"
#include <commondefs.h>
#include <coremsg.h>

struct core_node_msg *node_msg_alloc(int msg_len);
void node_msg_free(struct core_node_msg *msg);

extern atomic64_t ntransaction_id;

#define NODE_TRANSACTION_ID_MAX	((1ULL <<  32) - 1)

static inline uint64_t 
node_transaction_id(void)
{
	uint64_t ret;

	ret = atomic64_inc_return(&ntransaction_id);
	if (!ret)
		ret = atomic64_inc_return(&ntransaction_id);
	return ret;
}

int node_send_msg(sock_t *sock, struct core_node_msg *raw);
int node_msg_discard_pages(sock_t *sock, int pg_count);
int node_msg_discard_data(sock_t *sock, int dxfer_len);
sock_t * __node_sock_alloc(void *priv);
void __node_sock_free(sock_t *sock);

int node_sock_write(sock_t *sock, struct core_node_msg *raw);
int node_sock_write_page(sock_t *sock, pagestruct_t *page, int dxfer_len);
int node_sock_write_pgdata_list(sock_t *sock, struct pgdata *pglist, int pglist_cnt, uint32_t size);
int node_sock_read(sock_t *sock, void *buf, int dxfer_len);
int node_sock_write_data(sock_t *sock, uint8_t *buffer, int dxfer_len);
void node_sock_mark_close(sock_t *sock);
void node_sock_finish(sock_t *sock);
void node_sock_add_to_free_list(sock_t *sock);
void node_remove(uint64_t node_id, char *sys_rid);
void node_sock_write_error(sock_t *sock);
void node_sock_read_error(sock_t *sock);

/* timeouts */
#define NODE_SOCK_TIMEOUT		30000
#define NODE_GET_SOCK_TIMEOUT		30000

void node_msg_compute_csum(struct core_node_msg *raw);
int node_msg_csum_valid(struct core_node_msg *raw);


int sock_state_error(sock_t *sock);
int sys_sock_has_read_data(sock_t *sys_sock);
int sys_sock_is_connected(sock_t *sys_sock);
int sys_sock_has_write_space(sock_t *sys_sock);
int sys_sock_read(sock_t *sys_sock, void *buf, int len);
void sys_sock_free(sock_t *sys_sock);
void sys_sock_close(sock_t *sys_sock, int linger);
int sys_sock_write_page(sock_t *sys_sock, pagestruct_t *page, int offset, int len);
int sys_sock_write(sock_t *sys_sock, void *buf, int len);
sock_t * __sys_sock_create(void *priv, int domain);
void sock_activate(sock_t *sys_sock);
void sys_sock_nopush(sock_t *sys_sock, int set);
sock_t * sys_sock_accept(sock_t *sys_sock, void *priv, int *error, uint32_t *ipaddr);
int sys_sock_bind(sock_t *sys_sock, uint32_t addr, uint16_t port);
int sys_sock_shutdown(sock_t *sys_sock);
int sys_sock_connect(sock_t *sys_sock, uint32_t addr, uint32_t local_addr, uint16_t port);
int sys_sock_connect_local(sock_t *sys_sock, char *path);
void node_sock_wait_for_write(sock_t *sock);
int node_sock_read_nofail(sock_t *sock, void *buf, int dxfer_len);
void node_sock_start(sock_t *sock);
void node_sock_end(sock_t *sock);
void put_comm_sock(sock_t *sys_sock);
sock_t * get_comm_sock(void);

static inline uint32_t
calc_csum32(uint8_t *buf, int len)
{
        int i;
        uint32_t csum = 0x33332222;

        for (i = 0; (i + 4) <= len ; i+=4)
        {
                uint32_t val = *((uint64_t *)(buf+i));
                csum ^= val;
        }
	while (i < len)
		csum ^= buf[i++];
        return (csum);
}
#endif
