#ifndef QUADSTOR_ISTATE_H_
#define QUADSTOR_ISTATE_H_

#ifndef MSG_SIMPLE_TASK
#define MSG_SIMPLE_TASK			0x20 /* O/O */ /* SPI3 Terminology */
#endif

#ifndef MSG_HEAD_OF_QUEUE_TASK
#define MSG_HEAD_OF_QUEUE_TASK		0x21 /* O/O */ /* SPI3 Terminology */
#endif

#ifndef MSG_ORDERED_TASK
#define MSG_ORDERED_TASK		0x22 /* O/O */ /* SPI3 Terminology */
#endif

#ifndef MSG_ACA_TASK
#define MSG_ACA_TASK			0x24 /* 0/0 */ /* SPI3 */
#endif

#define AERC_MASK		0x80
#define TRMTSK_MASK		0x40
#define NORM_ACA_MASK		0x20
#define HISUP_MASK		0x10

#ifdef TAG_CMD_DEBUG
static inline void
debug_tag(struct qsio_scsiio *ctio)
{
	uint32_t id = (ticks % 3);

	switch (id) {
		case 0:
			id = MSG_HEAD_OF_QUEUE_TASK;
			break;
		case 1:
			id = MSG_ORDERED_TASK;
			break;
		default:
			id = MSG_SIMPLE_TASK;
			break;
	}
	ctio->task_attr = id;
}
#endif

static inline int 
istate_queue_cmd(struct initiator_state *istate, struct qsio_scsiio *ctio, int node_master)
{
	int exec = 0;

	istate_mtx_lock(istate->istate_lock);
	while (!node_master && atomic_read(&istate->blocked)) {
		istate_mtx_unlock(istate->istate_lock);
		wait_on_chan(istate->istate_wait, !atomic_read(&istate->blocked));
		istate_mtx_lock(istate->istate_lock);
	}

	if (ctio->task_attr == MSG_SIMPLE_TASK) {
		if (!istate->ordered && !istate->head) {
			ctio->queued = 1; /* ACA would mean that this is blocked however */ 
			exec = 1;
		}
	}
	else if (ctio->task_attr == MSG_HEAD_OF_QUEUE_TASK) {
		ctio->queued = 1; /* ACA would mean that this is blocked however */ 
		istate->head++;
		exec = 1;
	}
	else {
		if (!istate->queued) {
			ctio->queued = 1; /* ACA would mean that this is blocked however */ 
			istate->ordered++;
			exec = 1;
		}
	}

	if (!exec)
		istate->pending++;
	else
		istate->queued++;
	TAILQ_INSERT_TAIL(&istate->queue_list, ctio, ta_list);
	istate_mtx_unlock(istate->istate_lock);
	return exec;
}

static inline void
istate_abort_task_set(struct initiator_state *istate)
{
	struct qsio_scsiio *iter;

	istate_mtx_lock(istate->istate_lock);
	if (TAILQ_EMPTY(&istate->queue_list)) {
		istate_mtx_unlock(istate->istate_lock);
		return;
	}

	atomic_set(&istate->blocked, 1);
	TAILQ_FOREACH(iter, &istate->queue_list, ta_list) {
		if (iter->ccb_h.flags & QSIO_IN_DEVQ)
			continue;
		iter->ccb_h.flags |= QSIO_CTIO_ABORTED;
	}
	istate_mtx_unlock(istate->istate_lock);
}
 
static inline void
istate_abort_tasks(struct initiator_state *istate, uint64_t i_prt[], uint64_t t_prt[], uint8_t init_int, int block)
{
	struct qsio_scsiio *iter;

	istate_mtx_lock(istate->istate_lock);
	if (TAILQ_EMPTY(&istate->queue_list)) {
		istate_mtx_unlock(istate->istate_lock);
		return;
	}

	if (block)
		atomic_set(&istate->blocked, 1);

	TAILQ_FOREACH(iter, &istate->queue_list, ta_list) {
		if (iter->ccb_h.flags & QSIO_IN_DEVQ)
			continue;
		iter->ccb_h.flags |= QSIO_CTIO_ABORTED;
		if (!port_equal(iter->i_prt, i_prt) || !port_equal(iter->t_prt, t_prt) || iter->init_int != init_int)
			iter->ccb_h.flags |= QSIO_SEND_ABORT_STATUS;
	}
	istate_mtx_unlock(istate->istate_lock);
}

static inline int
istate_task_exists(struct initiator_state *istate, uint32_t task_tag)
{
	struct qsio_scsiio *iter;
	int task_found = 0;

	istate_mtx_lock(istate->istate_lock);
	TAILQ_FOREACH(iter, &istate->queue_list, ta_list) {
		if (iter->task_tag != task_tag)
			continue;
		task_found = 1;
		break;
	}
	istate_mtx_unlock(istate->istate_lock);
	return task_found;
}

static inline int
istate_abort_task(struct initiator_state *istate, uint64_t i_prt[], uint64_t t_prt[], uint8_t init_int, uint32_t task_tag, int *task_exists)
{
	struct qsio_scsiio *iter;
	int task_found = 0;

	debug_info("aborting task for tag %x\n", task_tag);
	istate_mtx_lock(istate->istate_lock);
	if (TAILQ_EMPTY(&istate->queue_list)) {
		istate_mtx_unlock(istate->istate_lock);
		return 0;
	}

	atomic_set(&istate->blocked, 1);
	TAILQ_FOREACH(iter, &istate->queue_list, ta_list) {
		debug_info("iter task tag %x task tag %x in devq %d\n", iter->task_tag, task_tag, iter->ccb_h.flags & QSIO_IN_DEVQ);
		if (iter->task_tag != task_tag)
			continue;

		if (iter->ccb_h.flags & QSIO_IN_DEVQ) {
			*task_exists = 1;
			break;
		}

		debug_info("aborting task with tag %x\n", task_tag);
		iter->ccb_h.flags |= QSIO_CTIO_ABORTED;
		if (!port_equal(iter->i_prt, i_prt) || !port_equal(iter->t_prt, t_prt) || iter->init_int != init_int)
			iter->ccb_h.flags |= QSIO_SEND_ABORT_STATUS;
		task_found = 1;
		break;
	}
	istate_mtx_unlock(istate->istate_lock);
	return task_found;
}
 
static inline void 
istate_remove_cmd(struct initiator_state *istate, struct qsio_scsiio *ctio, struct ccb_list *task_list)
{
	struct qsio_scsiio *iter;
	int done = 0;
	uint32_t ordered = 0, queued = 0;

	istate_mtx_lock(istate->istate_lock);
	if (ctio->queued) {
		istate->queued--;
		switch (ctio->task_attr) {
			case MSG_HEAD_OF_QUEUE_TASK:
				istate->head--;
				break;
			case MSG_ORDERED_TASK:
			case MSG_ACA_TASK:
				istate->ordered--;
				break;
		}
	}

	if (!istate->pending || (TAILQ_FIRST(&istate->queue_list) != ctio)) { /* Head of queue */
		TAILQ_REMOVE(&istate->queue_list, ctio, ta_list);
		if (TAILQ_EMPTY(&istate->queue_list))
			istate_chan_wakeup(istate->istate_wait);
		istate_mtx_unlock(istate->istate_lock);
		return;
	}

	TAILQ_REMOVE(&istate->queue_list, ctio, ta_list);

	TAILQ_FOREACH(iter, &istate->queue_list, ta_list) {
		if (iter->queued) {
			queued++;
			if (iter->task_attr != MSG_SIMPLE_TASK)
				ordered++;
			continue;
		}

		switch (iter->task_attr) {
			case MSG_SIMPLE_TASK:
				if (ordered) {
					done = 1;
					break;
				}
				iter->queued = 1;
				istate->pending--;
				istate->queued++;
				queued++;
				STAILQ_INSERT_TAIL(task_list, (struct qsio_hdr *)iter, c_list);
				break;
			case MSG_HEAD_OF_QUEUE_TASK:
				done = 1;
				break;
			case MSG_ORDERED_TASK:
			case MSG_ACA_TASK:
			default:
				if (queued) {
					done = 1;
					break;
				}
				iter->queued = 1;
				istate->pending--;
				istate->queued++;
				istate->ordered++;
				ordered++;
				queued++;
				STAILQ_INSERT_TAIL(task_list, (struct qsio_hdr *)iter, c_list);
				break;
		}
		if (done)
			break;
	}
	istate_mtx_unlock(istate->istate_lock);
}

extern uma_t *istate_cache;
static inline void
istate_free(struct initiator_state *istate)
{
	istate_mtx_free(istate->istate_lock);
	istate_wait_chan_free(istate->istate_wait);
	istate_uma_zfree(istate_cache, istate);
}

static inline void
istate_put(struct initiator_state *istate)
{
	if (atomic_dec_and_test(&istate->refs))
		istate_free(istate);
}

static inline void
istate_get(struct initiator_state *istate)
{
	atomic_inc(&istate->refs);
}

static inline void
init_istate(struct initiator_state *iter, uint64_t i_prt[], uint64_t t_prt[], uint16_t r_prt, uint8_t init_int)
{
	SLIST_INIT(&iter->sense_list);
	TAILQ_INIT(&iter->queue_list);
	istate_mtx_alloc(iter->istate_lock, "istate lock");
	istate_wait_chan_alloc(iter->istate_wait, "istate wait");
	port_fill(iter->i_prt, i_prt);
	port_fill(iter->t_prt, t_prt);
	iter->r_prt = r_prt;
	iter->init_int = init_int;
	atomic_set(&iter->refs, 1);
}

void istate_sense_list_free(struct initiator_state *istate);

static inline void
free_initiator_state(struct initiator_state *istate)
{
	wait_on_chan(istate->istate_wait, TAILQ_EMPTY(&istate->queue_list));
	istate_sense_list_free(istate);
	istate_put(istate);
}

static inline void
device_free_initiator_state2(struct tdisk *tdisk, uint64_t i_prt[], uint64_t t_prt[], uint8_t init_int)
{
	struct initiator_state *iter, *prev = NULL;

	tdisk_reservation_lock(tdisk);
	SLIST_FOREACH(iter, &tdisk->istate_list, i_list) {
		if (port_equal(iter->i_prt, i_prt) && port_equal(iter->t_prt, t_prt) && iter->init_int == init_int) {
			if (prev)
				SLIST_REMOVE_AFTER(prev, i_list);
			else
				SLIST_REMOVE_HEAD(&tdisk->istate_list, i_list);
			break;
		}
		prev = iter;
	}
	tdisk_reservation_unlock(tdisk);

	if (!iter)
		return;

	free_initiator_state(iter);
}

static inline void
device_remove_ctio(struct qsio_scsiio *ctio, struct ccb_list *ctio_list)
{
	struct initiator_state *istate;

	if (!ctio->ta_list.tqe_next && !ctio->ta_list.tqe_prev)
		return;

	istate = ctio->istate;
	if (!istate)
		return;

	istate_remove_cmd(istate, ctio, ctio_list);
	ctio->ta_list.tqe_next = NULL;
	ctio->ta_list.tqe_prev = NULL;
}

int fc_initiator_check(uint64_t wwpn[], uint64_t target_wwpn, void *device);

static inline struct initiator_state *
__device_get_initiator_state(struct tdisk *tdisk, uint64_t i_prt[], uint64_t t_prt[], uint16_t r_prt, uint8_t init_int, int alloc, int check)
{
	struct istate_list *istate_list = &tdisk->istate_list;
	struct initiator_state *iter, *prev = NULL;

	SLIST_FOREACH(iter, istate_list, i_list) {
		if (port_equal(iter->i_prt, i_prt) && port_equal(iter->t_prt, t_prt) && iter->init_int == init_int) {
			if (iter->disallowed)
				return NULL;
			if (prev) {
				SLIST_REMOVE_AFTER(prev, i_list);
				SLIST_INSERT_HEAD(istate_list, iter, i_list);
			}
			iter->timestamp = ticks;
			istate_get(iter);
			return iter;
		}
		prev = iter;
	}

	if (!alloc)
		return NULL;

	if (check && init_int == TARGET_INT_FC) {
		int allow;

		allow = fc_initiator_check(i_prt, t_prt[0], tdisk);
		if (allow != FC_RULE_ALLOW)
			return NULL;
	}

	iter = __uma_zalloc(istate_cache, Q_WAITOK);
	init_istate(iter, i_prt, t_prt, r_prt, init_int);
	iter->timestamp = ticks; 
	SLIST_INSERT_HEAD(istate_list, iter, i_list);
	istate_get(iter);
	return iter;
}

static inline struct initiator_state *
device_get_initiator_state(struct tdisk *tdisk, uint64_t i_prt[], uint64_t t_prt[], uint16_t r_prt, uint8_t init_int, int alloc, int check)
{
	return __device_get_initiator_state(tdisk, i_prt, t_prt, r_prt, init_int, alloc, check);
}


static inline int
__device_istate_queue_ctio(struct tdisk *tdisk, struct qsio_scsiio *ctio, int node_master)
{
	struct initiator_state *istate;
	int exec;

	tdisk_reservation_lock(tdisk);
	istate = device_get_initiator_state(tdisk, ctio->i_prt, ctio->t_prt, ctio->r_prt, ctio->init_int, 1, 1);
	tdisk_reservation_unlock(tdisk);
	if (!istate)
		return -1;
	ctio->istate = istate;
	exec = istate_queue_cmd(istate, ctio, node_master);
	return exec;
}

static inline int
device_istate_queue_ctio(struct tdisk *tdisk, struct qsio_scsiio *ctio)
{
	return __device_istate_queue_ctio(tdisk, ctio, 0);
}
#endif
