#ifndef QS_ITFDEFS_H_
#define QS_ITFDEFS_H_
#include "linuxdefs.h"
#include "exportdefs.h"
#include "queue.h"
#include <asm/ioctls.h>
#include <linux/namei.h>
#include <linux/vmalloc.h>
#include <commondefs.h>

#define debug_warn DEBUG_WARN_NEW
#define debug_print DEBUG_INFO_NEW
#define debug_check DEBUG_BUG_ON
#define debug_info(fmt,args...)	do {} while (0)
#define M_CTIODATA	0
#define M_PGLIST	0
#define M_PAGE_LIST	0
#define M_NODE_RMSG	0
#define M_MMAP		0
#define M_QUADSTOR	0
#define M_COMP		0
#define M_FILE		0
#define M_SOCK		0
#define ticks		jiffies
#define mtx_init	spin_lock_init

extern atomic_t itf_enabled;
extern atomic_t kern_inited;

static inline unsigned long
get_elapsed(unsigned long timestamp)
{
	unsigned long cur_ticks = jiffies;
	unsigned long elapsed;

	if (cur_ticks >= timestamp)
		elapsed = (cur_ticks - timestamp);
	else
		elapsed =  (timestamp - cur_ticks);
	return jiffies_to_msecs(elapsed);
}

/* queue helpers */
#define LIST_REMOVE_INIT(elm, field) do {			\
	if (elm->field.le_prev || elm->field.le_next) {		\
		LIST_REMOVE(elm, field);			\
		elm->field.le_prev = NULL;			\
		elm->field.le_next = NULL;			\
	}							\
} while (0)

#define TAILQ_REMOVE_INIT(head, elm, field) do {		\
	if (elm->field.tqe_prev || elm->field.tqe_next) {	\
		TAILQ_REMOVE(head, elm, field);			\
		elm->field.tqe_prev = NULL;			\
		elm->field.tqe_next = NULL;			\
	}							\
} while (0)

#define TAILQ_ENTRY_EMPTY(elm, field) 	((elm->field.tqe_prev == NULL && elm->field.tqe_next == NULL))
#define LIST_ENTRY_EMPTY(elm, field) 	((elm->field.le_prev == NULL && elm->field.le_next == NULL))

pagestruct_t* vm_pg_alloc(allocflags_t aflags);
void vm_pg_free(pagestruct_t *pp);
void* vm_pg_address(pagestruct_t *pp);
void vm_pg_ref(pagestruct_t *pp);
void vm_pg_unref(pagestruct_t *pp);
void __qs_free(void *ptr, int flags);
void* __qs_malloc(size_t size, int type, allocflags_t aflags);
void* __qszalloc(size_t size, int type, allocflags_t aflags);
void __uma_zfree(uma_t *cachep, void *ptr);
void * __uma_zalloc(uma_t *cachep, allocflags_t aflags);
void pgdata_free_page(struct pgdata *pgdata);
int pgdata_alloc_page(struct pgdata *pgdata, allocflags_t flags);
void pgdata_copy_page_ref(struct pgdata *dest, pagestruct_t *page);
void pgdata_add_page_ref(struct pgdata *dest, pagestruct_t *page);
void pgdata_copy_ref(struct pgdata *dest, struct pgdata *src);
void pgdata_add_ref(struct pgdata *dest, struct pgdata *src);
void pgdata_move(struct pgdata *dest, struct pgdata *src);
void pgdata_free(struct pgdata *pgdata);
struct pgdata * pgdata_allocate(uint32_t block_size, uint32_t num_blocks, int *ret_pglist_cnt, allocflags_t flags, int alloc_page);
void pglist_free(struct pgdata *pglist, int pglist_cnt);
struct qsio_scsiio * ctio_new(allocflags_t flags);
void ctio_allocate_buffer(struct qsio_scsiio *ctio, int length, allocflags_t flags);
void ctio_free_data(struct qsio_scsiio *ctio);
void ctio_free_data_nopage(struct qsio_scsiio *ctio);
void ctio_free_all(struct qsio_scsiio *ctio);
void __wait_complete_all(wait_compl_t *comp);
void __wait_for_done(wait_compl_t *comp);
long __wait_for_done_timeout(wait_compl_t *comp, long timo);

static inline void
chan_lock(wait_chan_t *chan)
{
	mtx_lock(&chan->mtx);
}

static inline void
chan_lock_intr(wait_chan_t *chan, void *data)
{
	unsigned long *flags = data;
	mtx_lock_irqsave(&chan->mtx, *flags);
}

static inline void
chan_unlock(wait_chan_t *chan)
{
	mtx_unlock(&chan->mtx);
}

static inline void
__chan_wakeup(wait_chan_t *chan)
{
	unsigned long flags;

	mtx_lock_irqsave(&chan->mtx, flags);
	wake_up_all(&chan->cv);
	mtx_unlock_irqrestore(&chan->mtx, flags);
}

static inline void
__chan_wakeup_one(wait_chan_t *chan)
{
	unsigned long flags;

	mtx_lock_irqsave(&chan->mtx, flags);
	wake_up(&chan->cv);
	mtx_unlock_irqrestore(&chan->mtx, flags);
}

static inline void
__chan_wakeup_one_unlocked(wait_chan_t *chan)
{
	wake_up(&chan->cv);
}

static inline void
__chan_wakeup_unlocked(wait_chan_t *chan)
{
	wake_up_all(&chan->cv);
}

static inline void
chan_unlock_intr(wait_chan_t *chan, void *data)
{
	unsigned long *flags = data;
	mtx_unlock_irqrestore(&chan->mtx, *flags);
}

int core_msg_send_ctio(struct qsio_scsiio *ctio);
int core_msg_send_ctio_mmap(struct qsio_scsiio *ctio);

#define	TYPE_BDEVICE	0x424C4F4B
#define	TYPE_TDEVICE	0x54444556
struct bdevice {
	int type;
	atomic_t disabled;
};

struct initiator_state {
	spinlock_t istate_lock;
	struct ctio_list queue_list;
	uint32_t ordered;
	uint32_t head;
	uint32_t queued;
	uint32_t pending;
	char name[256];
	uint64_t i_prt[2];
	uint64_t t_prt[2];
	uint16_t r_prt;
	uint8_t init_int;
	uint8_t disallowed;
	atomic_t blocked;
	atomic_t refs;
	unsigned long timestamp;
	SLIST_ENTRY(initiator_state) i_list;
	SLIST_HEAD(, sense_info) sense_list;
	wait_chan_t istate_wait;
};

SLIST_HEAD(istate_list, initiator_state);

struct tdisk {
	int type;
	int lun_id;
	int iscsi_tid;
	int vhba_id;
	int lba_shift;
	unsigned long flags;
	void *hpriv;
	atomic_t disabled;
	atomic_t pending_msg;
	TAILQ_HEAD(, qsio_scsiio) ctio_list;
	pagestruct_t *ctio_page; /* MMap page address */
	struct completion ctio_completion;
	spinlock_t tdevice_lock;
	wait_queue_head_t wait;
	unsigned long ctio_dxfer_len;
	int ctio_count;
	atomic_t refs;
	sx_t ctio_lock;
	spinlock_t ctio_alloc_lock;
	struct tdrive_stats stats;
	struct mutex reservation_lock;
	struct istate_list istate_list;
};

void tdevice_set_msg_pending(struct tdisk *tdevice);

struct mmap_entry {
	struct page *page;
	unsigned long start_addr;
	unsigned long end_addr;
	TAILQ_ENTRY(mmap_entry) e_list;
};
TAILQ_HEAD(mmap_entry_list, mmap_entry);

pagestruct_t * mmap_lookup_page(unsigned long addr, struct mmap_entry **ret);

#ifndef BYTE_ORDER
#define BYTE_ORDER 1234
#endif

#ifndef BIG_ENDIAN
#define BIG_ENDIAN 4321
#endif

#endif
