#ifndef QSIO_CCB_H_
#define QSIO_CCB_H_

#include <compilerdefs.h>
#ifndef FREEBSD
#include "queue.h"

enum {
	Q_WAITOK	= 0x1,
	Q_NOWAIT	= 0x2,
	Q_NOWAIT_INTR	= 0x4,
	Q_ZERO   	= 0x8,
};
#endif

enum {
	QS_IO_READ,
	QS_IO_WRITE,
	QS_IO_SYNC,
	QS_IO_SYNC_FLUSH,
};

static inline int
is_write_iop(int cmd)
{
	switch (cmd) {
	case QS_IO_READ:
		return 0;
	default:
		return 1;
	}
}

enum {
	QSIO_DIR_IN		= 0x01,
	QSIO_DIR_OUT		= 0x02,
	QSIO_DIR_NONE		= 0x04,
	QSIO_DATA_DIR_IN	= 0x08,
	QSIO_DATA_DIR_OUT	= 0x10,
	QSIO_SEND_STATUS	= 0x20,
	QSIO_TYPE_CTIO		= 0x40,
	QSIO_TYPE_NOTIFY	= 0x80,
	QSIO_HBA_ERROR		= 0x100,
	QSIO_HBA_REQUEUE	= 0x200,
	QSIO_CTIO_ABORTED	= 0x400,
	QSIO_SEND_ABORT_STATUS	= 0x800,
	QSIO_CLUSTERED		= 0x1000,
	QSIO_MIRROR		= 0x2000,
	QSIO_SYNC		= 0x4000,
	QSIO_IN_DEVQ		= 0x8000,
	QSIO_LOCKED		= 0x10000,
	QSIO_XCOPY		= 0x20000,
	QSIO_BUFFERED		= 0x40000,
	QSIO_DONE_SEND_STATUS	= 0x80000,
	QSIO_USR_PAGES		= 0x100000,
	QSIO_FREE_INIT_NAME	= 0x200000,
};

PACKED(struct qs_sense_data {
	uint8_t error_code;
	uint8_t segment;
	uint8_t flags;
	uint32_t info;
	uint8_t extra_len;
	uint8_t cmd_spec_info[4];
	uint8_t add_sense_code;
	uint8_t add_sense_code_qual;
	uint8_t fru;
	uint8_t sense_key_spec[3];
	uint8_t extra_bytes[46];
});

#define SENSE_LEN(sense_data)	((sense_data->extra_len + offsetof(struct qs_sense_data, cmd_spec_info)))


struct qpriv {
	void *qcmd;
	void *fcbridge;
	void *fcq;
};

struct iscsi_cmnd;
struct iscsi_priv {
	struct iscsi_cmnd *cmnd;
	char *init_name;
};

struct node_priv {
	void *node_msg;
	void *client_cond;
	void *client_lock;
	int  *client_done;
	int clientfd;
};

#ifdef FREEBSD 
struct vhba_priv {
	void *ccb;
};
#else
struct vhba_priv {
	struct scsi_cmnd *SCpnt;
	void (*done) (struct scsi_cmnd *);
};
#endif

struct ioctl_priv {
	void *completion;
};

struct qsio_hdr {
	void (*queue_fn) (void *);
	void *tdevice;
	uint32_t target_lun;
	uint32_t flags;	
	union {
		struct qpriv qpriv;
		struct iscsi_priv ipriv;
		struct vhba_priv vpriv;
		struct node_priv npriv;
		struct ioctl_priv cpriv;
	} priv;
	STAILQ_ENTRY(qsio_hdr) c_list;
};

STAILQ_HEAD(ccb_list, qsio_hdr);

enum {
	NOTIFY_STATUS_OK	= 0x00,
	NOTIFY_STATUS_ERROR	= 0x01,
};

#define INITIATOR_LOGOUT	0xFF02

struct qsio_immed_notify {
	struct    qsio_hdr ccb_h;
	uint8_t  init_int;
	uint8_t  notify_status;         /* Returned NOTIFY status */
	uint16_t r_prt;
	uint32_t fn;
	uint64_t i_prt[2];
	uint64_t t_prt[2];
	uint32_t task_tag;
};

enum {
	TASK_TYPE_SIMPLE	= 0,
	TASK_TYPE_HEAD		= 1,
	TASK_TYPE_ORDERED	= 2,
};

struct qsio_scsiio {
	struct qsio_hdr ccb_h;
	uint8_t  *sense_data;
	uint8_t  cdb[16];
	uint8_t  scsi_status;
	uint8_t  task_attr;
	uint8_t  queued:1;
	uint8_t  init_int:3;
	uint8_t  local_pool:1;
	uint8_t  pad1:3;
	uint8_t  sense_len;
	uint16_t r_prt;
	uint16_t pglist_cnt;
	uint8_t  *data_ptr;
	uint32_t dxfer_len;
	uint32_t task_tag;
	uint64_t i_prt[2];
	uint64_t t_prt[2];
	void     *istate;
	uint64_t tx_id;
	struct node_priv npriv;
	TAILQ_ENTRY(qsio_scsiio) hx_list;
	TAILQ_ENTRY(qsio_scsiio) ta_list;
	TAILQ_ENTRY(qsio_scsiio) ma_list;
	TAILQ_HEAD(, qsio_scsiio) ctio_list;
	uint32_t num_blocks;
	uint64_t start_timestamp;
	uint64_t queue_timestamp;
	uint64_t allocate_buffers_ticks;
	uint64_t dxfer_ticks;
	uint64_t response_timestamp;
	uint64_t end_timestamp;
};

TAILQ_HEAD(ctio_list, qsio_scsiio);

#endif
