/* 
 * Copyright (C) Shivaram Upadhyayula <shivaram.u@quadstor.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301, USA.
 */

#ifndef QUADSTOR_FCQ_H_
#define QUADSTOR_FCQ_H_

#include "qla_def.h"
#include "qla_target.h"

struct fc_cmd {
	int type;
	int need_free;
	int notify_fn;
	int data_dir;
	void *tgt_cmd;
	void *ccb;
	void *fcbridge;
	uint16_t cdb[16];
	uint32_t orig_length;
	uint32_t task_attr;
	uint64_t jiffies_at_send;
	TAILQ_ENTRY(fc_cmd) q_list;
	TAILQ_ENTRY(fc_cmd) f_list;
};

TAILQ_HEAD(fc_cmd_list, fc_cmd);

struct fcbridge {
	struct scsi_qla_host *ha;
	struct fcq *fcq;
	TAILQ_ENTRY(fcbridge) b_list;
	__u32 id;
	volatile int target_mode;
	volatile int initiator_mode;
};

struct fcq {
	unsigned long flags;
	spinlock_t fcq_lock;
	wait_queue_head_t fcq_wait;
	TAILQ_HEAD(, fc_cmd) pending_queue;
	struct task_struct *task;
	struct fcbridge *fcbridge;
};

struct fc_sess {
	TAILQ_ENTRY(fc_sess) s_list;
	TAILQ_ENTRY(fc_sess) a_list;
	struct fc_port *sess;
	spinlock_t sess_lock;
	atomic_t pending_cmds;
	uint32_t key;
	uint16_t loop_id;
	uint16_t active;
	uint8_t wwpn[WWN_SIZE];
	struct fc_cmd_list cmd_list;
};
TAILQ_HEAD(fc_sess_list, fc_sess);

struct lport {
	struct scsi_qla_host *vha;
	uint64_t wwpn;
	uint32_t id;
	struct fcbridge *fcbridge;
	TAILQ_ENTRY(lport) l_list;
	spinlock_t lport_lock; 
	struct fc_sess_list sess_list;
	struct fc_sess_list active_sess_list;
	struct fc_sess *loop_sess_list[65536]; /*loop id is uint16_t */
};
TAILQ_HEAD(lport_list, lport);

static inline void
fcq_insert_cmd(struct fcq *fcq, struct fc_cmd *cmd)
{
	unsigned long flags = 0;

	spin_lock_irqsave(&fcq->fcq_lock, flags);
	TAILQ_INSERT_TAIL(&fcq->pending_queue, cmd, q_list);
	spin_unlock_irqrestore(&fcq->fcq_lock, flags);
	wake_up(&fcq->fcq_wait);
}

enum {
	QLA_HDR_TYPE_CTIO = 13333,
	QLA_HDR_TYPE_NOTIFY = 25555,
};

struct fcbridge* fcbridge_new(void *ha, uint32_t id);
void fcbridge_exit(struct fcbridge *fcbridge);
void fcq_threads_exit(void);

struct fcq * fcq_init(char *namefmt, int id);
void fcq_exit(struct fcq *fcq);
struct fcq * fcq_get(struct fcbridge *fcbridge, uint64_t lun); 
int fcbridge_proc_cmd(void *bridge, void *iop);
int fcbridge_task_mgmt(struct fcbridge *fcbridge, struct qsio_immed_notify *notify);
void __ctio_free_data(struct qsio_scsiio *ctio);
void fcbridge_route_cmd_post(struct qsio_scsiio *ctio);
void fcbridge_insert_cmd(struct qsio_scsiio *ctio);
int fcbridge_i_prt_valid(struct fcbridge *fcbridge, uint64_t i_prt[]);

#endif
